import tkinter as tk
import tkinter.font as font
from tkinter import ttk
#from pathlib import Path
#from PIL import Image, ImageTk

from thonny.codeview import CodeViewText
from thonny.shell import ShellText
from thonny import get_workbench, get_shell


from .common import common_postit_tabs, common_images

class BaseWidget(ttk.Frame):

    def widget_init(self, tab_name):
        """ 
            frame structure(2 row sub-frame)

            main_frame (postit_button + main_note)
            bottom_frame(bottom_note)

            either main_note or bottom_note according to long_note flag  
        """
        # store tab
        self.tab_name = tab_name
        self.tab = common_postit_tabs[tab_name]
        

        # image
        self.enter_image = common_images['enter_small']

        # frame init
        ttk.Frame.__init__(self, self.tab.frame)

        # main and bottom sub-frame
        self.main_frame = ttk.Frame(self)
        self.main_frame.pack(side=tk.TOP)
        self.bottom_frame = ttk.Frame(self)
        self.bottom_frame.pack(side=tk.TOP)
        
        # button and label
        f = font.Font(size=11, weight=font.NORMAL, family='Consolas')
        self.postit_button = tk.Button(self.main_frame,  
                                        relief='flat',
                                        borderwidth=0,
                                        text='***' , 
                                        fg=self.tab.font_color, 
                                        bg=self.tab.fill_color,
                                        justify='left', 
                                        font=f,
                                        compound='right',
                                        #image=self.enter_image,
                                        padx=5,
                                        pady=5, 
                                        state='normal',
                                        )

        self.main_note_label = ttk.Label(self.main_frame, text='' ) 
        self.bottom_note_label = ttk.Label(self.bottom_frame, text='')

        # 1st row sub-frame
        self.postit_button.pack(side=tk.LEFT, anchor='w')
        self.main_note_label.pack(side=tk.LEFT, anchor='w',padx=5)
        # 2nd row sub-frame
        
        self.bottom_note_label.pack(side=tk.TOP, anchor='center',padx=5)



class BaseCode:
    def code_init(self, code, code_display=None, note=None, postfix_enter=None,
                     long_note=False):
        # tk var
        self.var_postfix_enter = tk.BooleanVar()
        self.var_postfix_enter.set(False)

        self.long_note = long_note

        if code_display is None:
            code_display = code
        self.code = code
        self.code_display = code_display 
        self.note = note
        if postfix_enter:
            self.var_postfix_enter.set(True)
        
        self.update_postit_code()

    def update_button_enter_sign(self):
        if self.var_postfix_enter.get():
            self.postit_button.config(image=self.enter_image)
        else:
            self.postit_button.config(image='')

    def toggle_postfix_enter(self):
        r = self.var_postfix_enter.get()
        self.var_postfix_enter.set(r)
        self.update_button_enter_sign()
        

    def set_code_display(self, text):
        self.postit_button.config(text=text)

    def set_code(self, text):
        pass

    def set_note(self, help_text):
        if not self.long_note:
            # use main note
            self.main_note_label.config(text=help_text)
            self.bottom_note_label.config(text='')
            self.bottom_note_label.pack_forget()
        else: # bottom note
            self.main_note_label.config(text='')
            self.bottom_note_label.pack(side=tk.LEFT,padx=15)
            self.bottom_note_label.config(text=help_text) 

    def update_postit_code(self):
        #self.set_code(self.code)
        self.set_code_display(self.code_display)
        self.set_note(self.note)
        self.update_button_enter_sign()

class BasePost:
    def post_init(self):
        self.drag_window = None
        self.drag_button = None
        self.drag_hover_selection = False
        self.hover_text_backup = ''
        #self.mouse_dragging = False
        # drag and press event
        self.postit_button.bind("<B1-Motion>", self.on_mouse_drag)
        self.postit_button.bind("<ButtonRelease-1>", self.on_mouse_release)
        self.postit_button.config(cursor='arrow')

    def on_mouse_drag(self, event):
        ###print('drag ...')
        #create drag window
        if not self.drag_window: 
            self.create_drag_window()
            self.postit_button.config(cursor='hand2')

        x_root, y_root = event.x_root, event.y_root

        self.drag_window.geometry('+{}+{}'.format(x_root+18, y_root+5))

        #change insert over editor or shell (but not postit button)
        
        hover_widget = event.widget.winfo_containing(x_root, y_root)
        
        if isinstance(hover_widget, CodeViewText):
            # hover editor
            editor_text = hover_widget
            relative_x = x_root - editor_text.winfo_rootx()
            relative_y =  y_root - editor_text.winfo_rooty()
            mouse_index = editor_text.index(f"@{relative_x},{relative_y}")
            # set cursor in editor
            editor_text.focus_set()
            editor_text.mark_set(tk.INSERT, mouse_index)

            if editor_text.tag_ranges(tk.SEL):
                #check darg hover selection
                if editor_text.compare(tk.SEL_FIRST, "<=", mouse_index) and \
                    editor_text.compare(mouse_index, "<=", tk.SEL_LAST):
                    self.drag_hover_selection = True
                    #self.drag_button.config(text='【取代】'+self.hover_text_backup)
                    
                else:
                    self.drag_hover_selection = False
                    #self.drag_button.config(text=self.hover_text_backup)
                    
                    
        elif isinstance(hover_widget, ShellText):
            # hover shell 
            shell_text = hover_widget
            relative_x = x_root - shell_text.winfo_rootx()
            relative_y = y_root - shell_text.winfo_rooty()
            # set cursor in shell
            shell_text.focus_set()
            mouse_index = shell_text.index(f"@{relative_x},{relative_y}")
            input_start_index = shell_text.index('input_start')
            if shell_text.compare(mouse_index, '>=', input_start_index):
                shell_text.mark_set(tk.INSERT, mouse_index)

                if shell_text.tag_ranges(tk.SEL):
                    #check darg hover selection
                    if shell_text.compare(tk.SEL_FIRST, "<=", mouse_index) and \
                        shell_text.compare(mouse_index, "<=", tk.SEL_LAST):
                        self.drag_hover_selection = True
                        #self.drag_button.config(text='【取代】'+self.hover_text_backup)
                        
                    else:
                        self.drag_hover_selection = False
                        #self.drag_button.config(text=self.hover_text_backup)
                        

    def create_drag_window(self):
            self.drag_window = tk.Toplevel()
            # clone postit_button in drag window
            image = self.postit_button.cget('image')
            compound = self.postit_button.cget('compound')
            font = self.postit_button.cget('font')
            bg = self.postit_button.cget('bg')
            fg = self.postit_button.cget('fg')
            text = self.postit_button.cget('text')
            #self.hover_text_backup = text
            justify = self.postit_button.cget('justify')
            self.drag_button = tk.Button(self.drag_window, text=text, bg=bg, 
                        fg=fg,font=font, compound=compound, image=image,
                        relief='solid', justify=justify, bd=0 )
            self.drag_button.pack()
            self.drag_window.overrideredirect(True)
            self.drag_window.attributes('-topmost', 'true')

    def on_mouse_release(self, event):
        #check and destroy drag window        
        if self.drag_window:
            self.drag_window.destroy()
            self.drag_window = None
            self.postit_button.config(cursor='arrow')
            
        # restore mouse cursor 

        # find out post type and target   
        x_root, y_root = event.x_root, event.y_root
        hover_widget = event.widget.winfo_containing(x_root,y_root)
        self.determine_post_place_and_type(hover_widget)
        #reset hover state 
        self.drag_hover_selection = False

    def determine_post_place_and_type(self, hover_widget):
        """
      post hover_widget:
        (1)postit_button : press button
        (2)editor_text : drag to editor
        (3)shell_text  : drag to shell

      post type:
        (1)press postit insert (pressing=True, selecting=False)
        (2)press postit insert with selection
                                (pressing=True, selecting=True)
        (3)drag postit insert (dragging=True, hovering=False)
        (4)drag postit insert hovering over selection
                                (dragging=True, hovering=True) 
         """        
        if hover_widget is self.postit_button: 
            # postit button pressed
            workbench = get_workbench()
            focus_widget = workbench.focus_get()
            if isinstance(focus_widget, CodeViewText):
                # cursor in editor
                editor_text = focus_widget 
                if editor_text.tag_ranges(tk.SEL)  :
                    # has selection
                    self.insert_into_editor(editor_text, 
                                            pressing=True, selecting=True)
                else:# no selection
                    self.insert_into_editor(editor_text, 
                                            pressing=True, selecting=False)
            elif isinstance(focus_widget, ShellText):
                # cusor in shell
                shell_text = focus_widget
                if shell_text.tag_ranges(tk.SEL):
                    # has selection
                    self.insert_into_shell(shell_text, 
                                           pressing=True, selecting=True)
                else:# no selection
                    self.insert_into_shell(shell_text, 
                                           pressing=True, selecting=False)
        elif isinstance(hover_widget, CodeViewText):
            # drag to editor 
            editor_text = hover_widget
            
            if self.drag_hover_selection:
                # drag_hover_selection
                self.insert_into_editor(editor_text, 
                                    dragging=True, hovering=True)
                
            else:# not drag_hover_selection
                self.insert_into_editor(editor_text, 
                                    dragging=True, hovering=False)
        elif isinstance(hover_widget, ShellText):
            # drag to shell
            shell_text = hover_widget
            
            if self.drag_hover_selection:
                # drag_hover_selection
                self.insert_into_shell(shell_text, 
                                    dragging=True, hovering=True)
            else:# no drag_hover_selection
                self.insert_into_shell(shell_text, 
                                    dragging=True, hovering=False)

    def insert_into_editor(self, editor_text, 
                           pressing=False, dragging=False,
                           selecting=False, hovering=False):
        if pressing and not selecting:
            self.content_insert(editor_text, self.code)
            if self.var_postfix_enter.get():
                editor_text.event_generate("<Return>")

        elif pressing and selecting:
            editor_text.event_generate("<BackSpace>")
            self.content_insert(editor_text, self.code)
            if self.var_postfix_enter.get():
                editor_text.event_generate("<Return>")

        elif dragging and not hovering:
            # cancel selection
 
            if editor_text.tag_ranges(tk.SEL):
                ori_sel_first = editor_text.index(tk.SEL_FIRST)
                ori_sel_last = editor_text.index(tk.SEL_LAST)
                editor_text.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
            
            self.content_insert(editor_text, self.code)
            if self.var_postfix_enter.get():
                editor_text.event_generate("<Return>")



        elif dragging and hovering:
            editor_text.event_generate("<BackSpace>")
            self.content_insert(editor_text, self.code)
            if self.var_postfix_enter.get():
                editor_text.event_generate("<Return>")

    def content_insert(self, text_widget, content):

        lines = content.split('\n')  
        #print(lines)
        line_num = len(lines)          
        if line_num == 1 :
            # one line (no newline)
                #print('co chi')
                text_widget.insert(tk.INSERT,lines[0])
        elif line_num > 1 :
            #multi lines 
            line_count = len(lines)
            for i, line in enumerate(lines):

                #if else else and except, need to add a extra backspack
                # pass is special case
                #if not line.startswith('pass'):
                if line[:4] in ('else', 'elif', 'exce')  :
                    text_widget.event_generate("<BackSpace>")

                text_widget.insert(tk.INSERT,line)

                #  generate enter if not last item
                if i < line_count - 1 :
                    text_widget.event_generate("<Return>")
                    



    def insert_into_shell(self, shell_text, 
                           pressing=False, dragging=False,
                           selecting=False, hovering=False):
        if pressing and not selecting:
            if shell_text.compare(tk.INSERT, '>=', 'input_start'):
                # cursor after input_start
                self.content_insert(shell_text, self.code)
                if self.var_postfix_enter.get():
                    shell_text.event_generate("<Return>")
            else: # cursor before input_start
                # append last line
                shell_text.mark_set(tk.INSERT, 'end-1c')
                self.content_insert(shell_text, self.code)
                if self.var_postfix_enter.get():
                    shell_text.event_generate("<Return>")
        elif pressing and selecting:
            if shell_text.compare(tk.SEL_LAST, '>', 'input_start'):
                if shell_text.compare(tk.SEL_FIRST, '>=', 'input_start'):
                    shell_text.delete(tk.SEL_FIRST, tk.SEL_LAST)
                    self.content_insert(shell_text, self.code)
                    if self.var_postfix_enter.get():
                        shell_text.event_generate("<Return>")
                else: # input_start among selection
                    shell_text.delete('input_start', tk.SEL_LAST)
                    shell_text.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
                    self.content_insert(shell_text, self.code)
                    if self.var_postfix_enter.get():
                        shell_text.event_generate("<Return>")
                
        elif dragging and not hovering:
            if shell_text.tag_ranges(tk.SEL):
                shell_text.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)

            if shell_text.compare(tk.INSERT, '>=', 'input_start'):
                # cursor after input_start
                self.content_insert(shell_text, self.code)
                if self.var_postfix_enter.get():
                    shell_text.event_generate("<Return>")

        elif dragging and hovering:
            if shell_text.compare(tk.SEL_LAST, '>', 'input_start'):
                if shell_text.compare(tk.SEL_FIRST, '>=', 'input_start'):
                    shell_text.delete(tk.SEL_FIRST, tk.SEL_LAST)
                    self.content_insert(shell_text, self.code)
                    if self.var_postfix_enter.get():
                        shell_text.event_generate("<Return>")
                else: # input_start among selection
                    shell_text.delete('input_start', tk.SEL_LAST)
                    shell_text.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
                    self.content_insert(shell_text, self.code)
                    if self.var_postfix_enter.get():
                        shell_text.event_generate("<Return>")

        


class BasePopup:
    def popup_init(self):
        self.popup_menu = tk.Menu(self, tearoff=0)
        self.popup_menu.add_checkbutton(label="【選項】結尾Enter換行", onvalue=1, offvalue=0, 
                variable=self.var_postfix_enter,
                command=self.toggle_postfix_enter,
                )
        self.postit_button.bind("<Button-3>", self.popup)

    def popup(self, event):
        self.popup_menu.tk_popup(event.x_root, event.y_root)



class BasePostit(BaseWidget, BaseCode, BasePost, BasePopup):
    """ composite and mixin approach base function postit"""
    def __init__(self,  tab_name, code, code_display=None, note=None, 
                    postfix_enter=False, long_note=False):
        self.widget_init(tab_name)
        self.code_init(code, code_display, note, postfix_enter, long_note)
        self.post_init()
        self.popup_init()
        


    



import tkinter as tk
from tkinter import ttk

from thonny.codeview import CodeViewText
from thonny.shell import ShellText
from thonny import get_workbench, get_shell

from ..base_postit import BaseCode, BasePost, BasePopup
from ..common import common_images

class ToolWidget(ttk.Frame):

    def widget_init(self, master, tool_name):
        # don't need to handle tab
        self.tool_name = tool_name
        self.tool_image = common_images[tool_name]

        ttk.Frame.__init__(self, master)        
        self.postit_button = tk.Button(self,  
                                        relief='groove',
                                        borderwidth=0,
                                        
                                        #fg=self.tab.font_color, 
                                        #bg=self.tab.fill_color,
                                        #justify='left', 
                                        #font=f,
                                        compound='right',
                                        image=self.tool_image,
                                        padx=0,
                                        pady=0, 
                                        )
        self.postit_button.pack(side=tk.LEFT, anchor='w')
        

        #self.note_label = ttk.Label(self, text='' )
        #self.note_label.pack(side=tk.LEFT, anchor='w',padx=5)

class ToolCodeMixin:
    def code_init(self):
        # tk var
        self.var_postfix_enter = tk.BooleanVar()
        self.var_postfix_enter.set(False)

        self.code_display = '' 
        self.note = ''
        self.code = ''
            
            #self.update_postit_code()

    def update_postit_code(self):
        pass


class ToolPostMixin:





    def insert_into_editor(self, text_widget, selecting, dragging):
        if self.tool_name == 'backspace':
            if not dragging:
                text_widget.event_generate("<BackSpace>")
            else: # dragging
                if selecting:
                    text_widget.event_generate("<BackSpace>")
                else:
                    text_widget.delete(tk.INSERT + '-1c')

        elif self.tool_name == 'undo':
            text_widget.edit_undo()
        elif self.tool_name == 'redo':
            text_widget.edit_redo()
        elif self.tool_name == 'enter':
            if not dragging:
                if selecting :
                    text_widget.event_generate("<BackSpace>")
                    text_widget.event_generate("<Return>")
                else : # not selecting
                    text_widget.event_generate("<Return>")
            else: # dragging
                if selecting:
                    text_widget.event_generate("<BackSpace>")
                    text_widget.event_generate("<Return>")
                else:
                    text_widget.insert(tk.INSERT, '\n')
                    #stored_index = text_widget.index(tk.INSERT)
                    #text_widget.tag_remove(tk.SEL, '1.0')
                    #text_widget.mark_set(tk.INSERT, stored_index)
                    #text_widget.event_generate("<Return>")
        elif self.tool_name == 'indent':
            if not dragging:
                text_widget.indent_region()
            else: # dragging
                if selecting:
                    text_widget.indent_region()
                else:
                    text_widget.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
                    text_widget.indent_region()

        elif self.tool_name == 'dedent':
            if not dragging:
                text_widget.dedent_region()
            else: # dragging
                if selecting:
                    text_widget.dedent_region()
                else:
                    text_widget.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
                    text_widget.dedent_region()


    def insert_into_shell(self, text_widget, selecting, dragging):
        if text_widget.compare(tk.INSERT, '>=' , 'input_start'): 
            if self.tool_name == 'backspace' and text_widget.compare(tk.INSERT, '>' , 'input_start'):
                # just bigger than, no equal than because of backspace del left char
                if not dragging:
                    if selecting:
                        if text_widget.compare(tk.SEL_FIRST, '>=', 'input_start'):
                            text_widget.event_generate("<BackSpace>")
                        elif text_widget.compare(tk.SEL_LAST, '>', 'input_start'):
                            #print('so co')
                            text_widget.delete('input_start', tk.SEL_LAST)
                            text_widget.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
                            text_widget.mark_set(tk.INSERT, 'input_start')
                    else:
                        text_widget.event_generate("<BackSpace>")
                else: # dragging
                    if selecting:
                        text_widget.event_generate("<BackSpace>")
                    else:
                        text_widget.delete(tk.INSERT + '-1c')                
            elif self.tool_name == 'undo':
                if text_widget.compare('input_start', '==','end-1c'):
                    # empty line
                    text_widget.event_generate("<Up>")
                else: # not empty line
                    text_widget.edit_undo()
            elif self.tool_name == 'redo':
                text_widget.edit_redo()
            elif self.tool_name == 'enter':
                if selecting:
                    text_widget.event_generate("<BackSpace>")
                    text_widget.event_generate("<Return>")
                else:# not selecting
                    text_widget.event_generate("<Return>")




            elif self.tool_name == 'indent':
                pass # when in shell
            elif self.tool_name == 'dedent':
                pass # when in shell

        else: # insert before input_start
            if self.tool_name == 'enter':
                text_widget.event_generate("<Return>")
            elif self.tool_name == 'backspace':
                # check if any selecting after input_start
                if text_widget.compare(tk.SEL_LAST, '>', 'input_start'):
                    text_widget.delete('input_start', tk.SEL_LAST)
                    text_widget.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
                    text_widget.mark_set(tk.INSERT, 'input_start')
            elif self.tool_name == 'undo':
                text_widget.event_generate('<Up>')
            elif self.tool_name == 'redo':
                text_widget.event_generate('<Down>')





class ToolPostit(ToolWidget, 
                 ToolCodeMixin, BaseCode,
                 ToolPostMixin, BasePost, 
                 BasePopup):
    """ composite and mixin approach postit"""
    def __init__(self, master, tool_name):
        self.widget_init(master, tool_name)
        self.code_init()
        self.post_init()
        #self.popup_init()


import tkinter as tk
import tkinter.font as font
from tkinter import ttk

from thonny.codeview import CodeViewText
from thonny.shell import ShellText
from thonny import get_workbench, get_shell

from .base_postit import BaseWidget,BaseCode, BasePost, BasePopup
from .common import common_postit_tabs,common_images

class EnclosedWidget(ttk.Frame):

    def widget_init(self, tab_name):
        # store tab
        self.tab_name = tab_name
        self.tab = common_postit_tabs[tab_name]
        
        # image
        self.enter_image = common_images['enter_small']

        # frame init
        ttk.Frame.__init__(self, self.tab.frame)

        # main and bottom sub-frame
        self.main_frame = ttk.Frame(self)
        self.main_frame.pack(side=tk.TOP)
        self.bottom_frame = ttk.Frame(self)
        self.bottom_frame.pack(side=tk.TOP)

        #f = font.Font(size=11, weight=font.NORMAL, family='Microsoft JhengHei')
        self.enclosed_left_image = common_images['enclosed_left']
        self.head_decoration_label = tk.Label(self.main_frame, text='',
                                    image=self.enclosed_left_image,
                                            )
        self.head_decoration_label.pack(side=tk.LEFT, anchor='w',padx=0)

        f = font.Font(size=11, weight=font.NORMAL, family='Consolas')
        self.postit_button = tk.Button(self.main_frame,  
                                        relief='flat',
                                        borderwidth=0,
                                        text='***' , 
                                        fg=self.tab.font_color, 
                                        bg=self.tab.fill_color,
                                        justify='left', 
                                        font=f,
                                        compound='right',
                                        #image=self.enter_image,

                                        )
        # 1st row sub-frame
        self.postit_button.pack(side=tk.LEFT, anchor='w')
        
        self.enclosed_right_image = common_images['enclosed_right']
        self.tail_decoration_label = tk.Label(self.main_frame, text='', 
                                            image= self.enclosed_right_image,)
        self.tail_decoration_label.pack(side=tk.LEFT, anchor='w',padx=0)

        self.main_note_label = ttk.Label(self.main_frame, text='' )
        self.main_note_label.pack(side=tk.LEFT, anchor='w',padx=5)
        # 2nd row sub-frame
        self.bottom_note_label = ttk.Label(self.bottom_frame, text='')
        self.bottom_note_label.pack(side=tk.TOP, anchor='center',padx=5)


class EnclosedCodeMixin:
    def code_init(self, enclosed_head, enclosed_tail, 
                        code_display, note, postfix_enter, long_note=False):
        # tk var
        self.var_postfix_enter = tk.BooleanVar()
        self.var_postfix_enter.set(False)

        self.long_note = long_note

        self.enclosed_head = enclosed_head
        self.enclosed_tail = enclosed_tail
        self.code = enclosed_head + enclosed_tail

        if not code_display :
            self.code_display = self.code
        else: 
            self.code_display = code_display
        self.note = note

        if postfix_enter:
            self.var_postfix_enter.set(True)
        
        self.update_postit_code()



class EnclosedPostMixin:
    def insert_into_editor(self, editor_text, 
                           pressing=False, dragging=False,
                           selecting=False, hovering=False):
        if pressing and not selecting:
            # insert enclosed and keep cursor between
            editor_text.insert(tk.INSERT, self.enclosed_head)
            stored_index = editor_text.index(tk.INSERT)
            editor_text.insert(tk.INSERT, self.enclosed_tail)
            #keep insert cursor in the stored_index 
            editor_text.mark_set(tk.INSERT, stored_index)
        elif pressing and selecting:
            # enclosed slection
            editor_text.insert(tk.SEL_FIRST, self.enclosed_head)
            stored_index = editor_text.index(tk.SEL_LAST)
            editor_text.insert(tk.SEL_LAST, self.enclosed_tail)
            #keep insert cursor in the stored_index 
            editor_text.mark_set(tk.INSERT, stored_index)
            editor_text.tag_remove(tk.SEL,'0.0', tk.END)
        elif dragging and not hovering:
            # insert enclosed and keep cursor between
            editor_text.insert(tk.INSERT, self.enclosed_head)
            stored_index = editor_text.index(tk.INSERT)
            editor_text.insert(tk.INSERT, self.enclosed_tail)
            #keep insert cursor in the stored_index 
            editor_text.mark_set(tk.INSERT, stored_index)
        elif dragging and hovering:
            # enclosed slection
            editor_text.insert(tk.SEL_FIRST, self.enclosed_head)
            stored_index = editor_text.index(tk.SEL_LAST)
            editor_text.insert(tk.SEL_LAST, self.enclosed_tail)
            #keep insert cursor in the stored_index 
            editor_text.mark_set(tk.INSERT, stored_index)
            editor_text.tag_remove(tk.SEL,'0.0', tk.END)

        if self.var_postfix_enter.get():
            text_widget.event_generate("<Return>")

        # if selecting:
        #     # enclosed slection
        #     text_widget.insert(tk.SEL_FIRST, self.enclosed_head)
        #     stored_index = text_widget.index(tk.SEL_LAST)
        #     text_widget.insert(tk.SEL_LAST, self.enclosed_tail)
        #     #keep insert cursor in the stored_index 
        #     text_widget.mark_set(tk.INSERT, stored_index)
        #     text_widget.tag_remove(tk.SEL,'0.0', tk.END)

        # else: # no selecting    
        #     text_widget.insert(tk.INSERT, self.enclosed_head)
        #     stored_index = text_widget.index(tk.INSERT)
        #     text_widget.insert(tk.INSERT, self.enclosed_tail)
        #     #keep insert cursor in the stored_index 
        #     text_widget.mark_set(tk.INSERT, stored_index)

        # if self.var_postfix_enter.get():
        #     text_widget.event_generate("<Return>")

    def insert_into_shell(self, shell_text, 
                           pressing=False, dragging=False,
                           selecting=False, hovering=False):

        if pressing and not selecting:
            if shell_text.compare(tk.INSERT, '>=' , 'input_start'):
                # just insert 
                shell_text.insert(tk.INSERT, self.enclosed_head)
                stored_index = shell_text.index(tk.INSERT)
                #keep insert cursor in the middle
                shell_text.insert(tk.INSERT, self.enclosed_tail)
                shell_text.mark_set(tk.INSERT, stored_index)

        elif pressing and selecting:
            # check selection after input_start
            if shell_text.compare(tk.SEL_LAST, '>', 'input_start'):
                if shell_text.compare(tk.SEL_FIRST, '>=', 'input_start'):
                    # enclosed selection(original )
                    shell_text.insert(tk.SEL_FIRST, self.enclosed_head)
                    stored_index = shell_text.index(tk.SEL_LAST)
                    #keep insert cursor in the last of selection
                    shell_text.insert(tk.SEL_LAST, self.enclosed_tail) 
                    shell_text.mark_set(tk.INSERT, stored_index)
                    shell_text.tag_remove(tk.SEL,tk.SEL_FIRST, tk.SEL_LAST)
                else: # enclosed selection(truncated )
                    shell_text.insert('input_start', self.enclosed_head)
                    stored_index = shell_text.index(tk.SEL_LAST)
                    #keep insert cursor in the last of selection
                    shell_text.insert(tk.SEL_LAST, self.enclosed_tail) 
                    shell_text.mark_set(tk.INSERT, stored_index)
                    shell_text.tag_remove(tk.SEL,tk.SEL_FIRST, tk.SEL_LAST)                    

        elif dragging and not hovering:
            if shell_text.compare(tk.INSERT, '>=' , 'input_start'):
                # just insert 
                shell_text.insert(tk.INSERT, self.enclosed_head)
                stored_index = shell_text.index(tk.INSERT)
                #keep insert cursor in the middle
                shell_text.insert(tk.INSERT, self.enclosed_tail)
                shell_text.mark_set(tk.INSERT, stored_index)

        elif dragging and hovering:
            # check selection after input_start
            if shell_text.compare(tk.SEL_LAST, '>', 'input_start'):
                if shell_text.compare(tk.SEL_FIRST, '>=', 'input_start'):
                    # enclosed selection(original )
                    shell_text.insert(tk.SEL_FIRST, self.enclosed_head)
                    stored_index = shell_text.index(tk.SEL_LAST)
                    #keep insert cursor in the last of selection
                    shell_text.insert(tk.SEL_LAST, self.enclosed_tail) 
                    shell_text.mark_set(tk.INSERT, stored_index)
                    shell_text.tag_remove(tk.SEL,tk.SEL_FIRST, tk.SEL_LAST)
                else: # enclosed selection(truncated )
                    shell_text.insert('input_start', self.enclosed_head)
                    stored_index = shell_text.index(tk.SEL_LAST)
                    #keep insert cursor in the last of selection
                    shell_text.insert(tk.SEL_LAST, self.enclosed_tail) 
                    shell_text.mark_set(tk.INSERT, stored_index)
                    shell_text.tag_remove(tk.SEL,tk.SEL_FIRST, tk.SEL_LAST)           
        
        if self.var_postfix_enter.get():
            text_widget.event_generate("<Return>")

        # if shell_text.compare(tk.INSERT, '>=' , 'input_start'):
        #     if selecting:
        #         # enclosed slection
        #         shell_text.insert(tk.SEL_FIRST, self.enclosed_head)
        #         stored_index = shell_text.index(tk.SEL_LAST)
        #         shell_text.insert(tk.SEL_LAST, self.enclosed_tail)
        #         #keep insert cursor in the stored_index 
        #         shell_text.mark_set(tk.INSERT, stored_index)
        #         shell_text.tag_remove(tk.SEL,'0.0', tk.END)

        #     else: # no selecting    
        #         shell_text.insert(tk.INSERT, self.enclosed_head)
        #         stored_index = shell_text.index(tk.INSERT)
        #         shell_text.insert(tk.INSERT, self.enclosed_tail)
        #         #keep insert cursor in the stored_index 
        #         shell_text.mark_set(tk.INSERT, stored_index)
        # else: # insert before input start
        #     if selecting:
        #         pass
        #     else: # not selecting
        #         shell_text.mark_set(tk.INSERT, 'end-1c')

        #         shell_text.insert(tk.INSERT, self.enclosed_head)
        #         stored_index = shell_text.index(tk.INSERT)
        #         shell_text.insert(tk.INSERT, self.enclosed_tail)
        #         #keep insert cursor in the stored_index 
        #         shell_text.mark_set(tk.INSERT, stored_index)






class EnclosedPostit(EnclosedWidget, 
                     EnclosedCodeMixin, BaseCode,
                     EnclosedPostMixin, BasePost, 
                     BasePopup):
    """  enclosed postit linke () [] '' "" {}  only for single line """
    def __init__(self, tab_name, enclosed_head, enclosed_tail,
                    code_display=None, note=None, postfix_enter=False,
                    long_note=False):
        self.widget_init(tab_name)
        # remove newline
        enclosed_head = enclosed_head.replace('\n','')
        enclosed_tail = enclosed_tail.replace('\n','')
        self.code_init(enclosed_head, enclosed_tail, 
                        code_display, note, postfix_enter, long_note)
        self.post_init()
        self.popup_init()




import tkinter as tk
from tkinter import ttk

from thonny.codeview import CodeViewText
from thonny.shell import ShellText
from thonny import get_workbench, get_shell

from ..base_postit import BaseCode, BasePost, BasePopup
from .tool_postit import ToolWidget, ToolCodeMixin
from ..common import common_images


class PilcrowPostMixin:

    def insert_into_editor(self, editor_text, 
                            pressing=False, dragging=False,
                            selecting=False, hovering=False):
        # get text_widget again for sure
        editor = get_workbench().get_editor_notebook().get_current_editor()
        code_text_widget = editor.get_text_widget()
        #code_text_widget.focus_set()
        y_scroll_pos = code_text_widget.yview()[0]
        x_scroll_pos = code_text_widget.xview()[0]

        #print("y view: ", code_text_widget.yview())
        #print('cursor  :', code_text_widget.cget('cursor'))
        #print('bg: ', self.postit_button.cget('bg'))

        if not self.show_pilcrow_mode:
            #  turning on  show_pilcrow_mode
            s = code_text_widget.get('1.0', 'end-1c')
            s = s.replace('\n', '¶\n')
            s = s.replace(' ', '·')
            code_text_widget.delete('1.0', 'end-1c')
            code_text_widget.insert('1.0', s)
                       
            code_text_widget.config(state=tk.DISABLED)
            code_text_widget.yview_moveto(y_scroll_pos)
            code_text_widget.xview_moveto(x_scroll_pos) 
            self.show_pilcrow_mode = True
            # highlight button
            self.postit_button.config(bg='orange')
            
            code_text_widget.config(cursor="left_side")
            #register event to failsafe back to read-write mode 
            code_text_widget.bind('<Button-1>', self.clickedWhenReadOnlyMode)
            #print("register failsafe event")


        else: # turning off  show_pilcrow_mode
            # unset read-only
            code_text_widget.config(state=tk.NORMAL)
            s = code_text_widget.get('1.0', 'end-1c')
            s = s.replace('¶\n', '\n')
            s = s.replace('·', ' ')
            code_text_widget.delete('1.0', 'end-1c')
            code_text_widget.insert('1.0', s)
            
            code_text_widget.yview_moveto(y_scroll_pos)
            code_text_widget.xview_moveto(x_scroll_pos)
            self.show_pilcrow_mode = False
            #restore button
            self.postit_button.config(bg='SystemButtonFace')
            
            code_text_widget.config(cursor="xterm")
            #unregister failsafe event   
            code_text_widget.unbind('<Button-1>')
            #print("cancel failsafe event")

    def insert_into_shell(self, shell_text, 
                            pressing=False, dragging=False,
                            selecting=False, hovering=False):
        if not dragging:
            self.insert_into_editor(shell_text)

    def clickedWhenReadOnlyMode(self, event):
        self.insert_into_editor(event.widget, selecting=False, dragging=False)


class PilcrowToolPostit(ToolWidget, 
                 ToolCodeMixin, BaseCode,
                 PilcrowPostMixin, BasePost, 
                 BasePopup):
    """ composite and mixin approach postit"""
    def __init__(self, master):
        self.show_pilcrow_mode = False

        self.widget_init(master, 'pilcrow')
        self.code_init()
        self.post_init()
        #self.popup_init()


import tkinter as tk

from thonny import tktextext 
from thonny.ui_utils import EnhancedTextWithLogging
from thonny import get_workbench, get_shell, get_runner
import thonny.codeview
from thonny.codeview import PythonText

# monkey patching
# replace with  selection event custom handler
class CodeViewText(EnhancedTextWithLogging, PythonText):
    """Provides opportunities for monkey-patching by plugins"""

    def __init__(self, master=None, cnf={}, **kw):

        if "replace_tabs" not in kw:
            kw["replace_tabs"] = False

        super().__init__(
            master=master,
            tag_current_line=get_workbench().get_option("view.highlight_current_line"),
            cnf=cnf,
            **kw
        )
        # Allow binding to events of all CodeView texts
        self.bindtags(self.bindtags() + ("CodeViewText",))
        tktextext.fixwordbreaks(tk._default_root)
        
        # monkey patching 
        self.bind('<<Selection>>', self.on_monkey, True)

    def on_monkey(self, event):
        sel = self.get(tk.SEL_FIRST, tk.SEL_LAST)
        print('monkey :', sel, " ,  len :", len(sel))

    def on_secondary_click(self, event=None):
        super().on_secondary_click(event)
        self.mark_set("insert", "@%d,%d" % (event.x, event.y))

        menu = get_workbench().get_menu("edit")
        try:
            from thonny.plugins.debugger import get_current_debugger

            debugger = get_current_debugger()
            if debugger is not None:
                menu = debugger.get_editor_context_menu()
        except ImportError:
            pass

        menu.tk_popup(event.x_root, event.y_root)

thonny.codeview.CodeViewText = CodeViewText
print('  monkey patched :  CodeViewText')


import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from collections import Counter
from keyword import iskeyword

import thonny.codeview
from thonny.codeview import CodeViewText
from thonny.shell import ShellText
from thonny import get_workbench, get_shell
from thonny.plugins.notes import NotesText

from .tool_postit import ToolWidget, ToolCodeMixin
from ..base_postit import BaseCode, BasePost, BasePopup

from ..common import  common_default_vars, common_images
from .. import common                       



class VariableMenuWidget(ttk.Frame):
    """ composite and mixin approach postit"""
    def widget_init(self, master, update_after_run=False):
        # just combobox. don't need to handle tab and image
        self.last_focus = ''
        self.vars_counter = None
        self.var_update_after_run = tk.BooleanVar()
        if update_after_run:
            self.var_update_after_run.set(True)
        else:
            self.var_update_after_run.set(False)

        self.vars_limit = 100
        self.tk_var = tk.StringVar()

        ttk.Frame.__init__(self, master)        
        self.vars_combobox = ttk.Combobox(self, width=10, state="readonly",
                justify=tk.CENTER,textvariable=self.tk_var,takefocus=0,
                values=[])
        self.restore_default_vars()

        self.vars_combobox.pack(side=tk.LEFT, anchor='w')

        #self.vars_combobox.current(0)

        self.vars_combobox.bind('<<ComboboxSelected>>', self.on_combo_select)
        #self.vars_combobox.bind('<<Selection>>', self.on_select)
        self.vars_combobox.bind('<Button-1>',self.on_combo_click)

        get_workbench().bind("ToplevelResponse", self._handle_toplevel_response, True)

    def empty_vars(self):
        self.vars_counter = Counter()
        self.update_vars_menu()

    def restore_default_vars(self):
        #if self.vars_counter:
        #    del self.vars_counter

        if common_default_vars:   
            self.vars_counter = Counter(common_default_vars)
            self.update_vars_menu()
        else:
            self.vars_counter = Counter()
            self.update_vars_menu()

    def update_vars_menu(self):
        if len(self.vars_counter):
            vars_list =[v for v,_ in self.vars_counter.most_common(self.vars_limit)]
            self.vars_combobox.config(values=vars_list)
            self.vars_combobox.current(0)
            #
            #print('here enable')
            common.enable_var_buttons()
        else:
            self.vars_combobox.config(values='')
            self.tk_var.set('')
            #
            common.disable_var_buttons()



    def on_combo_select(self, event):
        if self.last_focus is not '':
            self.last_focus.focus_set()
            
        self.selection_clear()

    def on_combo_click(self, event):
  
        workbench = get_workbench()
        self.last_focus = workbench.focus_get()
        #self.selection_clear()
        
    def _handle_toplevel_response(self, event):
        #print('got toplevel event')
        if self.var_update_after_run.get():
            if "globals" in event:
                #self.vars_counter.update(event['globals'].keys())
                # only add var the first time
                for key in event['globals'].keys():
                    if key not in self.vars_counter:
                        self.vars_counter[key] = 1
                self.update_vars_menu()

class VariableMenuPopup:
    def popup_init(self):
        self.popup_menu = tk.Menu(self, tearoff=0)

        self.popup_menu.add_command(
            label="刪除目前變數", command=self.delete_current)
        self.popup_menu.add_command(
            label="恢復預設變數", command=self.delete_all_restore_default)
        self.popup_menu.add_command(
            label="清空全部變數", command=self.delete_all)
        self.popup_menu.add_separator()
        self.popup_menu.add_checkbutton(label="【選項】執行後更新變數(全域)",
                onvalue=1, offvalue=0, 
                variable=self.var_update_after_run,
                )

        self.vars_combobox.bind("<Button-3>", self.popup)

    def popup(self, event):
        try:
            self.popup_menu.tk_popup(event.x_root, event.y_root)
        finally:
            self.popup_menu.grab_release()

    def delete_current(self):
        var = self.tk_var.get()
        if var:
            del self.vars_counter[var]
            self.update_vars_menu()

    def delete_all_restore_default(self):
        
        ans = messagebox.askyesno('刪除變數','刪除所有變數，並恢復預設值嗎？')
        #print(ans)
        if ans:
            #print('yes ')
            self.restore_default_vars()
        else: # no
            return

    def delete_all(self):
        if len(self.vars_counter):
            ans = messagebox.askyesno('清空全部變數','要清空全部變數嗎？')
            #print(ans)
            if ans:
                #print('yes ')
                self.empty_vars()
                
            else: # no
                return

class VariableMenuPostit(VariableMenuWidget, 
                    VariableMenuPopup
                 ):
    """ composite and mixin approach postit"""
    def __init__(self, master, update_after_run=False):
        self.widget_init(master, update_after_run)
        #self.code_init()
        #self.post_init()
        self.popup_init()


        

class VariableAddToolPostit(ttk.Frame):
    """special postit"""
    def __init__(self, master):
        # don't need to handle tab
        self.tool_name = 'variable_add'
        self.tool_image = common_images[self.tool_name]
        self.select_text = ''

        ttk.Frame.__init__(self, master)        
        self.postit_button = tk.Button(self,  
                                        relief='groove',
                                        borderwidth=0,
                                        compound='right',
                                        image=self.tool_image,
                                        padx=0,
                                        pady=0,
                                        state='disable',
                                        command=self.on_mouse_click
                                        )
        self.postit_button.pack(side=tk.LEFT, anchor='w')
        
        # bind text select on code view , shell and notes 
        self.bind_all('<<Selection>>',self.on_text_select,True)

    def on_text_select(self, event):
        widget = event.widget
        if isinstance(widget,CodeViewText) or isinstance(widget,ShellText)\
                                    or isinstance(widget, NotesText):
                        
            # only select within a line
            self.select_text = widget.get(tk.SEL_FIRST,tk.SEL_LAST)
            if len(self.select_text) and '\n' not in self.select_text:
                #print(self.select_text)
                self.postit_button.config(state='normal')
            else:
                self.select_text = ''
                self.postit_button.config(state='disable')
                

    def on_mouse_click(self):
        #print("clicked")
        if not self.select_text.isidentifier():
            content = '【 ' + self.select_text + ' 】 不是一個合格的變數名稱\n\n'
            content += '【說明】1.變數名稱可以用的字是文字,底線(_)或數字\n'
            content += '　　　　2.變數名稱的開頭第1個字不可以用數字'
            messagebox.showwarning('變數名稱錯誤', content)
            return
        elif iskeyword(self.select_text):
            content = '【 ' + self.select_text + ' 】 是python的保留關鍵字\n'
            content += '不適合用來作為變數名稱\n'
            content += '請修改或是換一個名稱'
            messagebox.showwarning('變數名稱錯誤', content)
            return
        else: # var name ok
            vars_postit = common.share_vars_postit
            #print(vars_postit)
            vars_postit.vars_counter[self.select_text] += 1 
            vars_postit.update_vars_menu()
            vars_postit.tk_var.set(self.select_text)



class VariableFetchToolPostMixin:
    def on_mouse_drag(self, event):
        state = self.postit_button.cget('state')
        if  state in ('normal', 'active') :
            BasePost.on_mouse_drag(self, event)
        else: # state is disable . do nothing
            pass

    def on_mouse_release(self, event):
        state = self.postit_button.cget('state')
        if state in ('normal', 'active') :
            BasePost.on_mouse_release(self, event)
            #print('here')
        else: # state is disable . do nothing
            #print('else')
            #print(self.postit_button.cget('state'))
            pass

    def create_drag_window(self):
        
        self.drag_window = tk.Toplevel()
        # get var text
        font = self.postit_button.cget('font')

        if self.tool_name == 'variable_get':
            text = common.share_vars_postit.tk_var.get() + ' '
        elif self.tool_name == 'variable_assign':
            text = common.share_vars_postit.tk_var.get() + ' = '
        elif self.tool_name == 'variable_plus_assign':
            text = common.share_vars_postit.tk_var.get() + ' += '
        elif self.tool_name == 'variable_minus_assign':
            text = common.share_vars_postit.tk_var.get() + ' -= '
        elif self.tool_name == 'variable_comma':
            text = common.share_vars_postit.tk_var.get() + ', '
        elif self.tool_name == 'variable_dot':
            text = common.share_vars_postit.tk_var.get() + '.'
        elif self.tool_name == 'variable_parentheses':
            text = common.share_vars_postit.tk_var.get() + '() '
        elif self.tool_name == 'variable_square':
            text = common.share_vars_postit.tk_var.get() + '[] '
        else:
            text = ''

        if self.tool_name == 'variable_get':
            color = '#7af85a'
        else:
            color = '#5acef8'

        self.drag_button = tk.Button(self.drag_window, text=text, bg=color, 
                    font=font, fg='black', relief='solid', bd=0 )
        self.drag_button.pack()
        self.drag_window.overrideredirect(True)
        self.drag_window.attributes('-topmost', 'true')

    def content_insert(self, text_widget, content):
        vars_postit = common.share_vars_postit
        var = vars_postit.tk_var.get()
        #var_content = var + ' '
        
        if self.tool_name == 'variable_get':
            var_content = var 
        elif self.tool_name == 'variable_assign':
            var_content = var + ' = '
        elif self.tool_name == 'variable_plus_assign':
            var_content = var + ' += '
        elif self.tool_name == 'variable_minus_assign':
            var_content = var + ' -= '
        elif self.tool_name == 'variable_comma':
            var_content = var + ', '
        elif self.tool_name == 'variable_dot':
            var_content = var + '.'
        elif self.tool_name == 'variable_parentheses':
            var_content = var  + '() '
        elif self.tool_name == 'variable_square':
            var_content = var + '[] '
        else:
            var_content = ''

        text_widget.insert(tk.INSERT, var_content)

        # adjust insert index
        if self.tool_name == 'variable_square':
            text_widget.mark_set(tk.INSERT, 'insert -2c')


        # add counter and  update menu according to most common
        vars_postit.vars_counter[var] += 1
        vars_postit.update_vars_menu()
        vars_postit.tk_var.set(var)


class VariableFetchToolPopup:
    def popup_init(self):
        self.popup_menu = tk.Menu(self, tearoff=0)
        self.popup_menu.add_command(label="V    取出變數",
            command=lambda:self.switch_button('variable_get'))
        self.popup_menu.add_command(label="V =  變數設值",
            command=lambda:self.switch_button('variable_assign'))
        self.popup_menu.add_command(label="V +=  加後設值",
            command=lambda:self.switch_button('variable_plus_assign'))
        self.popup_menu.add_command(label="V -=  減後設值",
            command=lambda:self.switch_button('variable_minus_assign'))
        self.popup_menu.add_command(label="V ,   變數逗號",
            command=lambda:self.switch_button('variable_comma'))
        self.popup_menu.add_command(label="V .   變數句點",
            command=lambda:self.switch_button('variable_dot'))
        self.popup_menu.add_command(label="V ( ) 變數圓括號",
            command=lambda:self.switch_button('variable_parentheses'))
        self.popup_menu.add_command(label="V [ ] 變數方括號",
            command=lambda:self.switch_button('variable_square'))


        self.postit_button.bind("<Button-3>", self.popup)

    def popup(self, event):
        #if self.tool_name != 'variable_get':
        self.popup_menu.tk_popup(event.x_root, event.y_root)
        

    def switch_button(self, tool_name):
        self.tool_name = tool_name
        self.tool_image = common_images[tool_name]
        self.postit_button.config(image=self.tool_image)

        # insert at the same time
        workbench = get_workbench()
        focus_widget = workbench.focus_get()
        if isinstance(focus_widget, CodeViewText):
            # cursor in editor
            editor_text = focus_widget 
            if editor_text.tag_ranges(tk.SEL)  :
                # has selection
                self.insert_into_editor(editor_text, 
                                        pressing=True, selecting=True)
            else:# no selection
                self.insert_into_editor(editor_text, 
                                        pressing=True, selecting=False)
        elif isinstance(focus_widget, ShellText):
            # cusor in shell
            shell_text = focus_widget
            if shell_text.tag_ranges(tk.SEL):
                # has selection
                self.insert_into_shell(shell_text, 
                                        pressing=True, selecting=True)
            else:# no selection
                self.insert_into_shell(shell_text, 
                                        pressing=True, selecting=False) 



class VariableFetchToolPostit(ToolWidget, 
                 ToolCodeMixin, BaseCode,
                 VariableFetchToolPostMixin, BasePost, 
                 VariableFetchToolPopup):
    """  used by 4 tools.
    variable_get, variable_assign, variable_comma, variable_dot"""

    def __init__(self, master, tool_name):
        self.widget_init(master, tool_name)
        self.code_init()
        self.post_init()
        self.popup_init()



# class VariableAssignToolPostMixin:
#     def create_drag_window(self):
#         self.drag_window = tk.Toplevel()
#         # get var text
#         font = self.postit_button.cget('font')
#         text = common.share_vars_postit.tk_var.get() + ' = '
#         self.drag_button = tk.Button(self.drag_window, text=text, bg='#7af85a', 
#                     font=font, fg='black', relief='solid', bd=0 )
#         self.drag_button.pack()
#         self.drag_window.overrideredirect(True)
#         self.drag_window.attributes('-topmost', 'true')

#     def content_insert(self, text_widget, content):
#         vars_postit = common.share_vars_postit
#         var = vars_postit.tk_var.get()
#         var_content = var + ' = '
#         text_widget.insert(tk.INSERT, var_content)

#         # add counter and  update menu according to most common
#         vars_postit.vars_counter[var] += 1
#         vars_postit.update_vars_menu()
#         vars_postit.tk_var.set(var)
#         #print(vars_postit.vars_counter)


# class VariableAssignToolPostit(ToolWidget, 
#                  ToolCodeMixin, BaseCode,
#                  VariableAssignToolPostMixin, BasePost, 
#                  BasePopup):
#     """ composite and mixin approach postit"""
#     def __init__(self, master):
#         self.widget_init(master, 'variable_assign')
#         self.code_init()
#         self.post_init()
#         #self.popup_init()


# class VariableCommaToolPostMixin:
#     def create_drag_window(self):
#         self.drag_window = tk.Toplevel()
#         # get var text
#         font = self.postit_button.cget('font')
#         text = common.share_vars_postit.tk_var.get() + ', '
#         self.drag_button = tk.Button(self.drag_window, text=text, bg='#7af85a', 
#                     font=font, fg='black', relief='solid', bd=0 )
#         self.drag_button.pack()
#         self.drag_window.overrideredirect(True)
#         self.drag_window.attributes('-topmost', 'true')

#     def content_insert(self, text_widget, content):
#         vars_postit = common.share_vars_postit
#         var = vars_postit.tk_var.get()
#         var_content = var + ', '
#         text_widget.insert(tk.INSERT, var_content)

#         # add counter and  update menu according to most common
#         vars_postit.vars_counter[var] += 1
#         vars_postit.update_vars_menu()
#         vars_postit.tk_var.set(var)


# class VariableCommaToolPostit(ToolWidget, 
#                  ToolCodeMixin, BaseCode,
#                  VariableCommaToolPostMixin, BasePost, 
#                  BasePopup):
#     """ composite and mixin approach postit"""
#     def __init__(self, master):
#         self.widget_init(master, 'variable_comma')
#         self.code_init()
#         self.post_init()
#         #self.popup_init()


# class VariableDotToolPostMixin:
#     def create_drag_window(self):
#         self.drag_window = tk.Toplevel()
#         # get var text
#         font = self.postit_button.cget('font')
#         text = common.share_vars_postit.tk_var.get() + '.'
#         self.drag_button = tk.Button(self.drag_window, text=text, bg='#7af85a', 
#                     font=font, fg='black', relief='solid', bd=0 )
#         self.drag_button.pack()
#         self.drag_window.overrideredirect(True)
#         self.drag_window.attributes('-topmost', 'true')

#     def content_insert(self, text_widget, content):
#         vars_postit = common.share_vars_postit
#         var = vars_postit.tk_var.get()
#         var_content = var + '.'
#         text_widget.insert(tk.INSERT, var_content)

#         # add counter and  update menu according to most common
#         vars_postit.vars_counter[var] += 1
#         vars_postit.update_vars_menu()
#         vars_postit.tk_var.set(var)


# class VariableDotToolPostit(ToolWidget, 
#                  ToolCodeMixin, BaseCode,
#                  VariableDotToolPostMixin, BasePost, 
#                  BasePopup):
#     """ composite and mixin approach postit"""
#     def __init__(self, master):
#         self.widget_init(master, 'variable_dot')
#         self.code_init()
#         self.post_init()
#         #self.popup_init()
